# =============================================================================
# <uniformity_analyzer to automatically evaluate PET image uniformity>
#     Copyright (C) <2021>  <Jaroslav Ptacek (WG PET/CT/MRI QC EFOMP)>
# 
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.
# =============================================================================

import os
from win32api import GetSystemMetrics

import pydicom as dcm

import tkinter as tk
import tkinter.filedialog as filedialog

from matplotlib import cm
from matplotlib.figure import Figure
from matplotlib.ticker import MaxNLocator, AutoMinorLocator
from matplotlib.ticker import FormatStrFormatter
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk

from PIL import Image, ImageTk

import numpy as np
import pandas as pd

# =============================================================================
# data object
# =============================================================================
class phantomData:
    def __init__(self, directory, data, reference_slice, size, all_setup):
# =============================================================================
# data object properties
# =============================================================================
        self.directory = directory
        self.data_original = data #(original data)
        self.data = [] #(working data)
        self.reference_slice = reference_slice # reference slice containing DICOM info
        self.size = size
        self.primary_zoom = []

        self.all_setup = all_setup

        self.zoom = 1

        self.uni_thresh = 5
        self.uni_shrink = 0
        self.uni_center = 5
        self.uni_end = 5

        self.position = [0,0] # phantom position setup rows x cols
        self.slice = int(size[2]/2) # initital slice to display pocatecni rez k otevreni
        self.ending_position = [0,self.size[2]]

        # nastaveni zpusobu zobrazeni - info[0] - s SUV / bez SUV
        # setting the display form - info[0] - with / without SUV
        # info[1] - threshold for outliers
        self.info = [0,0]

        # storage for calculated values from particular slices
        self.slice_SUVmean = []
        self.slice_SUVmax = []
        self.slice_SUVmin = []
        self.slice_diff_from_1 = []
        self.slice_COV = []

# =============================================================================
# data object functions
# =============================================================================
    def show_me_slices(self, of_what, which, where, position, zoom, SUV, uni_thresh, show_hide):
        self.primary_zoom = [self.size[0]/where.winfo_height(), self.size[1]/where.winfo_width()] # size is rows x cols
        temp = Image.fromarray(cm.Greys(of_what[:,:,which]/np.max(of_what[:,:,int(0.5*(self.ending_position[1]+self.ending_position[0]))]), bytes=True))
        temp = temp.resize((int(where.winfo_width()*zoom), int(where.winfo_height()*zoom))) # resize according to the window size cols x rows
        self.image_to_show = ImageTk.PhotoImage(temp)
        where.create_image(position[0],position[1], anchor='nw', image=self.image_to_show) # this is cols x rows

        if show_hide == 1:
            if SUV != 0:
                #  display of colour masks - red mask
                temp = of_what[:,:,which]*self.mask[:,:,which]
                temp[of_what[:,:,which]*self.mask[:,:,which] < (1+uni_thresh)*SUV] = 0 # not show
                temp[of_what[:,:,which]*self.mask[:,:,which] > (1+uni_thresh)*SUV] = 1 # show
                temp = temp*self.SUV_mask[:,:,which] # get rid of zero values


                temp = np.uint8(cm.Reds(temp, bytes=True))
                # colour channels mixing
                temp[...,3][temp[...,1] == 245] = 0
                temp[...,0][temp[...,3] == 255] = 255 # red
                temp[...,1]=0 # green
                temp[...,2]=0 # blue

                temp = Image.fromarray(temp.astype(np.uint8))
                temp = temp.resize((int(where.winfo_width()*zoom), int(where.winfo_height()*zoom))) # resize according to the window size cols x rows
                self.red_mask = ImageTk.PhotoImage(temp)
                where.create_image(position[0],position[1], anchor='nw', image=self.red_mask) # this is cols x rows

                #  display of colour masks - blue mask
                temp = []
                temp = of_what[:,:,which]*self.mask[:,:,which]
                temp[of_what[:,:,which]*self.mask[:,:,which] > (1-uni_thresh)*SUV] = 0
                temp[of_what[:,:,which]*self.mask[:,:,which] < (1-uni_thresh)*SUV] = 1
                temp = temp*self.SUV_mask[:,:,which] #get rid of zero valuest

                temp = np.uint8(cm.Blues(temp, bytes=True))

                # colour channels mixing
                temp[...,3][temp[...,1] == 251] = 0
                temp[...,2][temp[...,3] == 255] = 255
                temp[...,0]=0
                temp[...,1]=0

                temp = Image.fromarray(temp.astype(np.uint8))
                temp = temp.resize((int(where.winfo_width()*zoom), int(where.winfo_height()*zoom))) # resize according to the window size cols x rows
                self.blue_mask = ImageTk.PhotoImage(temp)
                where.create_image(position[0],position[1], anchor='nw', image=self.blue_mask) # this is cols x rows
            else:
                ...
        else:
            ...

    def slice_zoom(self, event, what, zoom_point, where, show_hide):
    # zoom[0] scroll button rotation, zoom[1] position row, zoom[2] position cols, where = whete to zoom
        zoom_step = 0.5
        if (zoom_point[0] > 0):
            if self.zoom >= 5: # setting the zoom = property
                ...
            else:
                self.zoom += (120/zoom_point[0])*zoom_step
                self.position = [self.position[0]-zoom_point[1]*zoom_step, self.position[1]-zoom_point[2]*zoom_step] # setting the position = property
                self.show_me_slices(what, self.slice, where, self.position, self.zoom, self.info[0], self.uni_thresh/100, show_hide)
        elif (zoom_point[0] < 0):
            if self.zoom <= 1:
                ...
            else:
                self.position = [self.position[0]+zoom_point[1]*zoom_step, self.position[1]+zoom_point[2]*zoom_step]
                self.zoom += (120/zoom_point[0])*zoom_step
                self.show_me_slices(what, self.slice, where, self.position, self.zoom, self.info[0], self.uni_thresh/100, show_hide)

    def slice_list(self, event, what, index, where, show_hide):
    # index = scroll button rotation, where = where to zoom
        self.slice += int(120/index) # setting the slice to display = property

        if self.slice <= 0:
            self.slice = 0

        elif self.slice >= self.size[2]-1:
            self.slice = self.size[2]-1

        self.show_me_slices(what, self.slice, where, self.position, self.zoom, self.info[0], self.uni_thresh/100, show_hide)

    def slice_in_SUV(self, what, volume, inj_activity, inj_time, inj_date):
        volume = float(volume)
        inj_activity = float(inj_activity)

        time_delta = (int(self.all_setup['series_time'][0:2])*3600+int(self.all_setup['series_time'][2:4])*60+int(self.all_setup['series_time'][4:6])) - \
        (int(inj_time[0:2])*3600+int(inj_time[2:4])*60+int(inj_time[4:6]))
          
        injected_activity_corrected = float(inj_activity)*np.exp(-np.log(2)*time_delta/float(self.all_setup['halflife']))
        injected_activity_corrected = injected_activity_corrected*1000*1000
        what = what/(injected_activity_corrected/volume)

        return(what)

    def slice_click_reversed(self, event, position): # from real to display coordinates
        position_r = position[0]*self.zoom/self.primary_zoom[0]+self.position[1]
        position_c = position[1]*self.zoom/self.primary_zoom[1]+self.position[0]
        position_s = position[2]
        return(position_r, position_c, position_s)

    def give_me_center(self, of_what, which, size):
        rows = list(range(0, size[0]))
        columns = list(range(0, size[1]))
        total_mass = np.sum(of_what[:,:,which])


        temp_row = columns*np.sum(of_what[:,:,which],0)
        temp_col = rows*np.sum(of_what[:,:,which],1)

        cm_row = float((sum(temp_row)/total_mass))
        cm_col = float((sum(temp_col)/total_mass))
        return(cm_row, cm_col)

    def find_me_phantom(self, what_mask, ROI_extent_shrink):
        temp = np.rot90(np.sum(what_mask[:,:,:], axis=0),3)
        temp = np.sum(temp,axis=1)
        temp = np.nonzero(temp)
        if ROI_extent_shrink != 0:
            end1 = temp[0][0]-self.uni_shrink
            end2 = temp[0][-1]+self.uni_shrink
        else:
            end1 = temp[0][0]
            end2 = temp[0][-1]
        return(end1, end2)

    def uni_mask(self, what, thresh):
        mask = np.zeros([what.size[0], what.size[1], what.size[2]])
        start = int(0.25*(what.ending_position[1]+what.ending_position[0]))
        end = int(0.75*(what.ending_position[1]+what.ending_position[0]))
        mask[what.data_original > thresh*np.max(what.data_original[:,:,start:end])] = 1

        return(mask)

    def give_me_SUV(self, what, thresh, diff):
        what.slice_SUVmean = []
        what.slice_SUVmax = []
        what.slice_SUVmin = []
        what.slice_diff_from_1 = []
        what.slice_COV = []
        what.slice_avolmean = []

        r = np.arange(0, what.size[0])
        c = np.arange(0, what.size[1])
        mask_array = np.zeros((what.size[0], what.size[1]))

        SUV_mask = np.zeros((what.size[0], what.size[1], what.size[2]))


        for i in range(self.ending_position[0], self.ending_position[1]+1):
            r_diam = float(np.count_nonzero(np.sum(what.mask[:,:,i], 0))+2*diff)
            c_diam = float(np.count_nonzero(np.sum(what.mask[:,:,i], 1))+2*diff)
            center_row, center_col = self.give_me_center(self.data, i, what.size)

            cc = center_row # position of the centre in row 
            cr = center_col # position of the centre in column

            a = float(c_diam/2) # axis a lenght
            b = float(r_diam/2) # axis b lenght
            
            # prevent division by 0
            if a == 0 or b == 0:
                a = 0.01
                b = 0.01

            mask = (r[:,np.newaxis]-cr)**2/a**2 + (c[np.newaxis,:]-cc)**2/b**2 < 1 # create an ellipse
            mask_array[mask] = 1

            SUV_mask[:,:,i] = mask_array
            temp = what.data[:,:,i][mask_array == 1]
            what.slice_SUVmean.append(np.mean(temp))
            what.slice_SUVmax.append(np.max(temp))
            what.slice_SUVmin.append(np.min(temp))
            what.slice_diff_from_1.append(np.mean(temp)*100-100)
            what.slice_COV.append(100*(np.std(temp)/np.mean(temp)))
            temp = what.data_original[:,:,i][mask_array == 1]
            what.slice_avolmean.append(np.mean(temp))
            # what.slice_COV.append(np.std(temp)) # COV = noise
            # what.slice_COV.append(np.mean(temp)) # COV = signal

        temp = self.data[SUV_mask == 1]

        mean_SUV = np.mean(temp)
        diff_from_1 = mean_SUV*100-100

        mean_SUV = round(mean_SUV, 4)
        diff_from_1 = round(diff_from_1, 2)
        
        return mean_SUV, diff_from_1, SUV_mask

    def get_out_results(self):
        temp = [['in_slice_SUVmean', 'slice_diff (%)', 'COV (%)', 'avol (kBq.cm-3)']]
        for i in range(0, len(self.slice_diff_from_1)):
            temp.append([self.slice_SUVmean[i], self.slice_diff_from_1[i], self.slice_COV[i], self.slice_avolmean[i]])

        file_name = 'SUV' + '_' + self.all_setup['device'] + '_' + self.all_setup['study_date'][0:2] + self.all_setup['study_date'][3:5] + \
            self.all_setup['study_date'][6:10] + '_' + 'diam_shrink_by_' + str(abs(self.uni_shrink)) + '_px_result.csv'
        file_name = file_name.replace(" ", "") # delete all spaces
        file_name = file_name.replace("/", "_") # delete all backslashes

        file_dir = filedialog.asksaveasfilename(parent=root, initialfile=file_name, initialdir="/",title='Please select a directory to save') # locate the data directory

        if file_dir: # check - any directory selected?
            pd.DataFrame(temp).to_csv(file_dir, sep=';')
            tk.messagebox.showinfo("Export", "Export finished!")
        else:
            ...

# =============================================================================
# graph creation object
# =============================================================================
class GraphDiffCOV:
    def __init__(self, which_frame, graph_size, graph_position, graph_diff, graph_COV):

        # dpi in dot per inch, graph_size in inches
        fig = Figure(figsize=(graph_size[0]/200, graph_size[1]/200), dpi = 200)
        axs = fig.subplots(2)
        fig.subplots_adjust(hspace=0.3, wspace = 0, top=0.95, bottom=0.15, left = 0.15, right=0.98)

        title_font = {'fontname':'Arial', 'size':'5'}
        axis_font = {'fontname':'Arial', 'size':'5'}
        axis_tick = {'labelsize':'4'}
        data_line = {'color':'red', 'marker':'', 'markersize':'2', 'linestyle':'-', 'linewidth':'0.75'}

        axs[0].plot(graph_diff[0],graph_diff[1], **data_line)
        axs[1].plot(graph_COV[0],graph_COV[1], **data_line)
        axs[0].set_title('per slice deviation', y=0.95, **title_font)
        axs[1].set_title('COV (%)', y=0.95, **title_font)
        axs[1].set_xlabel('slice #', labelpad=2, **axis_font)

        for i in range(0, len(axs)):
            axs[i].autoscale(enable=True)
            if i == 0:
                axs[i].set_ylabel('diff (%)',**axis_font)

            else:
                axs[i].set_ylabel('COV (%)',**axis_font)

            axs[i].tick_params(axis = 'both', **axis_tick)
            axs[i].xaxis.set_major_locator(MaxNLocator(integer=True))
            axs[i].xaxis.set_major_formatter(FormatStrFormatter('%.0f'))
            axs[i].xaxis.set_minor_locator(AutoMinorLocator())
            axs[i].yaxis.set_minor_locator(AutoMinorLocator())

            axs[i].grid(which = 'major', color = 'grey', linestyle = '-', linewidth = '0.25')
            axs[i].grid(which = 'minor', color = 'grey', linestyle = '--', linewidth = '0.20')

        axs[0].axes.xaxis.set_ticklabels([])
        axs[1].tick_params(axis='both', which='major', pad=1)

        self.canvas = FigureCanvasTkAgg(fig, master = which_frame)
        # navigation toolbar
        frame = tk.Frame(which_frame)
        frame.grid(row = graph_position[0]+1, column=0)
        toolbar = NavigationToolbar2Tk(self.canvas, frame).update()
        self.canvas.get_tk_widget().grid(row = graph_position[0], column = graph_position[1])

# =============================================================================
# mainWindow object
# =============================================================================
class mainWindow:
    def __init__(self, master):
        self.master = master

        # self.master.geometry('+1915+0') # optional to setup where the window opens - can be used for dual monitors
        self.master.title("Uniformity analyzer")
        self.master.resizable(0,0)

        self.Xpix = GetSystemMetrics(0)
        self.Ypix = GetSystemMetrics(1)
        self.Xsize = (self.Xpix-350)/3
        self.Ysize = self.Xsize

        self.plus = tk.PhotoImage(file = r"plus.png")
        self.minus = tk.PhotoImage(file = r"minus.png")

        self.ROI_extent_shrink = tk.IntVar()
        self.ROI_results_show = tk.IntVar()

        self.info_length = '0'
        self.info_diameter = '0'
        self.info_SUV = tk.StringVar()
        self.info_SUV_deviation = tk.StringVar()
        self.slice_infoLU = tk.StringVar()
        self.slice_infoLB = tk.StringVar()
        self.slice_infoRB = tk.StringVar()

        self.uni_graph = 0 # setup of the number of slices from frond/end to omit in graph

        self.frame_a = tk.LabelFrame(master, text='Data')
        self.frame_b = tk.LabelFrame(master, text='Per slice deviation and COV')
        self.frame_c = tk.LabelFrame(master, text='Load data, DICOM info')
        self.frame_d = tk.LabelFrame(master, text='Uniformity check settings')

        # =============================================================================
        # frame_a
        # =============================================================================
        self.phantom_slice_view = tk.Canvas(self.frame_a, width=self.Xsize, height=self.Ysize, borderwidth=2, bg='white')
        self.phantom_slice_view.config(highlightbackground='black', highlightthickness='2')
        self.phantom_slice_view.grid(row=0, column=0)

        label = tk.Label(self.frame_a, textvariable=self.slice_infoLU, anchor='ne', justify='left').grid(row=0, column=0, sticky='nw')
        self.slice_infoLU.set('slice: ???' + '\n' + 'mean SUV: ???' + '\n' + 'diff SUV (%): ???' + '\n' + 'COV (%): ???')
        label = tk.Label(self.frame_a, textvariable=self.slice_infoLB, anchor='ne', justify='left').grid(row=0, column=0, sticky='sw')
        self.slice_infoLB.set('SUV max: ???' + '\n' + 'SUV min: ???')
        label = tk.Label(self.frame_a, textvariable=self.slice_infoRB, anchor='ne', justify='left').grid(row=0, column=0, sticky='se')
        self.slice_infoRB.set('ROI diameter: 000.0' + ' mm' +  '\n' + 'VOI length: 000.0' + ' mm')

        # =============================================================================
        # frame_b
        # =============================================================================
        # initital graph display
        GraphDiffCOV(self.frame_b, (self.Xsize, self.Ysize), [0,0], [[],[]], [[],[]])

        # frame_c
        # =============================================================================
        button = tk.Button(self.frame_c, text="Open data", command=self.open_data)
        button.grid(row=0, column=0, sticky='w')
        self.data_size = tk.Label(self.frame_c, text='\n\n\n\n\n\n\n\n\n\n\n', anchor='sw', justify='left')
        self.data_size.grid(row=1, rowspan=20, column=0)

        label = tk.Label(self.frame_c, text='Phantom volume: ').grid(row=1, column=1, sticky = 'nw')
        self.data_volume = tk.Entry(self.frame_c, width = 8, exportselection=0)
        self.data_volume.grid(row=1, column=2, sticky = 'n')
        self.data_volume.insert(0, '6300')
        label = tk.Label(self.frame_c, text=' ml').grid(row=1, column=3, sticky = 'nw')

        label = tk.Label(self.frame_c, text='Phantom activity: ').grid(row=2, column=1, sticky = 'nw')
        self.data_activity = tk.Entry(self.frame_c, width = 8, exportselection=0)
        self.data_activity.grid(row=2, column=2, sticky = 'n')
        self.data_activity.insert(0, '0.000')
        label = tk.Label(self.frame_c, text=' MBq').grid(row=2, column=3, sticky = 'nw')

        label = tk.Label(self.frame_c, text='Phantom injection time: ').grid(row=3, column=1, sticky = 'nw')
        self.data_time = tk.Entry(self.frame_c, width = 8, exportselection=0)
        self.data_time.grid(row=3, column=2, sticky = 'n')
        self.data_time.insert(0, '000000')
        label = tk.Label(self.frame_c, text=' hhmmss').grid(row=3, column=3, sticky = 'nw')
        
        label = tk.Label(self.frame_c, text='Phantom injection date: ').grid(row=4, column=1, sticky = 'nw')
        self.data_date = tk.Entry(self.frame_c, width = 8, exportselection=0)
        self.data_date.grid(row=4, column=2, sticky = 'n')
        self.data_date.insert(0, '000000')
        label = tk.Label(self.frame_c, text=' ddmmyyyy').grid(row=4, column=3, sticky = 'nw')

        # =============================================================================
        # frame_d
        # =============================================================================

        label = tk.Label(self.frame_d, text='Change ROI diameter by: ', anchor='sw', justify='left').grid(row=1, column=0, sticky='nsew')
        self.uni_shrink_lbl = tk.Label(self.frame_d, width = 1)
        self.uni_shrink_lbl.grid(row=1, column=1, sticky='nsew')
        self.uni_shrink_lbl['text'] = '0.0 / 0'
        label = tk.Label(self.frame_d, text=' mm/px', anchor='sw', justify='left').grid(row=1, column=2, sticky='nsew')

        button = tk.Button(self.frame_d, image = self.minus, width="20", height="20", command=lambda: self.uni_shrink(-1)).grid(row=1, column=3, sticky='nsew')
        button = tk.Button(self.frame_d, image = self.plus, width="20", height="20", command=lambda: self.uni_shrink(1)).grid(row=1, column=4, sticky='nsew')

        ch_button = tk.Checkbutton(self.frame_d, text = "Change also the ROI extent by the same amount of pixels?", variable = self.ROI_extent_shrink, onvalue = 1, offvalue = 0, command= lambda i=0: self.uni_end(i)).grid(row=2, column = 0, columnspan=2, sticky='w')

        label = tk.Label(self.frame_d, text='Show outliers above -(blue) / +(red): ', anchor='sw', justify='left').grid(row=3, column=0, sticky='nsew')
        self.uni_thresh_lbl = tk.Label(self.frame_d, width = 1)
        self.uni_thresh_lbl.grid(row=3, column=1, sticky='nsew')
        self.uni_thresh_lbl['text'] = '???'
        label = tk.Label(self.frame_d, text=' %', anchor='sw', justify='left').grid(row=3, column=2, sticky='nsew')

        button = tk.Button(self.frame_d, image = self.minus, width="20", height="20", command=lambda: self.uni_thresh(-1)).grid(row=3, column=3, sticky='nsew')
        button = tk.Button(self.frame_d, image = self.plus, width="20", height="20", command=lambda: self.uni_thresh(1)).grid(row=3, column=4, sticky='nsew')

        ch_button = tk.Checkbutton(self.frame_d, text = "Show / hide color coding?", variable = self.ROI_results_show, onvalue = 1, offvalue = 0, command= self.ROI_results_hide).grid(row=4, column = 0, columnspan=2, sticky='w')
        self.ROI_results_show.set(1)

        label = tk.Label(self.frame_d, text='Phantom center threshold at : ', anchor='sw', justify='left').grid(row=5, column=0, sticky='nsew')
        self.uni_center_lbl = tk.Label(self.frame_d, width = 1)
        self.uni_center_lbl.grid(row=5, column=1, sticky='nsew')
        self.uni_center_lbl['text'] = '50'
        label = tk.Label(self.frame_d, text=' %', anchor='sw', justify='left').grid(row=5, column=2, sticky='nsew')

        button = tk.Button(self.frame_d, image = self.minus, width="20", height="20", command=lambda: self.uni_center(-1)).grid(row=5, column=3, sticky='nsew')
        button = tk.Button(self.frame_d, image = self.plus, width="20", height="20", command=lambda: self.uni_center(1)).grid(row=5, column=4, sticky='nsew')

        label = tk.Label(self.frame_d, text='Phantom end positions threshold at : ', anchor='sw', justify='left').grid(row=6, column=0, sticky='nsew')
        self.uni_end_lbl = tk.Label(self.frame_d, width = 1)
        self.uni_end_lbl.grid(row=6, column=1, sticky='nsew')
        self.uni_end_lbl['text'] = '???'
        label = tk.Label(self.frame_d, text=' %', anchor='sw', justify='left').grid(row=6, column=2, sticky='nsew')

        button = tk.Button(self.frame_d, image = self.minus, width="20", height="20", command=lambda: self.uni_end(-1)).grid(row=6, column=3, sticky='nsew')
        button = tk.Button(self.frame_d, image = self.plus, width="20", height="20", command=lambda: self.uni_end(1)).grid(row=6, column=4, sticky='nsew')

        label = tk.Label(self.frame_d, text='Calculated mean SUV: ', anchor='sw', justify='left').grid(row=7, column=0, sticky='nsew')
        label = tk.Label(self.frame_d, width = 1, textvariable=self.info_SUV)
        label.grid(row=7, column=1, sticky='nsew')
        label['text'] = '???'

        label = tk.Label(self.frame_d, text='Deviation from expected SUV=1: ', anchor='sw', justify='left').grid(row=8, column=0, sticky='nsew')
        label = tk.Label(self.frame_d, width = 1, textvariable=self.info_SUV_deviation)
        label.grid(row=8, column=1, sticky='nsew')
        label['text'] = '???'
        label = tk.Label(self.frame_d, text=' %', anchor='sw', justify='left').grid(row=8, column=2, sticky='nsew')

        label = tk.Label(self.frame_d, text='Crop data to graph by: ', anchor='sw', justify='left').grid(row=9, column=0, sticky='nsew')
        self.uni_graph_crop_lbl = tk.Label(self.frame_d, width = 1)
        self.uni_graph_crop_lbl.grid(row=9, column=1, sticky='nsew')
        self.uni_graph_crop_lbl['text'] = '0'
        label = tk.Label(self.frame_d, text=' slices', anchor='sw', justify='left').grid(row=9, column=2, sticky='nsew')

        button = tk.Button(self.frame_d, image = self.minus, width="20", height="20", command=lambda: self.uni_graph_crop(-1)).grid(row=9, column=3, sticky='nsew')
        button = tk.Button(self.frame_d, image = self.plus, width="20", height="20", command=lambda: self.uni_graph_crop(1)).grid(row=9, column=4, sticky='nsew')

        button = tk.Button(self.frame_d, text='Calculate results', command = self.calculate_SUV).grid(row=10, column=0)
        button = tk.Button(self.frame_d, text='Export results', command = self.get_out_results).grid(row=10, column=1)

        # =============================================================================
        # master.grid
        # =============================================================================
        self.frame_a.grid(row=0,rowspan=2, column=0, columnspan=2)
        self.frame_b.grid(row=0,rowspan=2, column=2, columnspan=2)
        self.frame_c.grid(row=2, column=0, columnspan=2, sticky='nsew')
        self.frame_d.grid(row=2, column=2, columnspan=2, sticky='nsew')
        frame_switch(self.frame_d, 0) # stays inactive until data loaded

    def open_data(self):

        dirname = filedialog.askdirectory(parent=root,initialdir="/",title='Please select a directory') # locate data directory

        # check if any directory selected
        if not dirname:
            return

        else:
            PT_lst = []  # create emty list

            for item in os.listdir(dirname): # lists the selected directory - only the selected one!!!
                if os.path.isfile(os.path.join(dirname,item)):
                    PT_lst.append(os.path.join(dirname,item))

        temp_data = {} # slice data stored in the dictionary

        # opens files one after another and stores them in the dictionary
        for i in range(0, len(PT_lst)):
            # check whether the file is DICOM
            try:
                temp = dcm.read_file(PT_lst[i]) # read the file with DICOM header and additional info
            except: # if file not DICOM - skipped, no action taken
                ...
                
            modality = str(temp[0x0008, 0x0060].value)
            number = int(temp[0x0020, 0x0013].value)
            position = int(temp[0x0020, 0x0032].value[2]) # sorting slices according to their position 
            # print(temp.DecayFactor)
                    
            # read the info from the reference file
            if modality == 'PT' and number == 1:
                ref_PT = dcm.read_file(PT_lst[i], stop_before_pixels=True)
                # print(ref_PT)
                
            # if calculated like this, with DecayCorrection = START, volume activity is always scaled by the PET itself to SeriesTime
            # SeriesTime is used to correct the injected activity for SUV calculations
            # volume activity at the time of measurement (which can differ from SeriesTime in dynamic studies) can be obtained by
            # dividing temp_data by a corresponding DecayFactor
            temp_data[position] = temp.pixel_array*temp.RescaleSlope + temp.RescaleIntercept
            
        data_info = {}

        # data_info as a dictionary
        data_info = {'rows':ref_PT[0x0028, 0x0010].value, 'cols':ref_PT[0x0028, 0x0011].value, 'slices':len(temp_data), \
                     'slice_thickness':str(ref_PT.SliceThickness), \
                     'pixel_size':str(ref_PT.PixelSpacing[0]), \
                     'study_date': ref_PT.StudyDate[6:8]+'.'+ref_PT.StudyDate[4:6]+'.'+ref_PT.StudyDate[0:4], \
                     'series_date': ref_PT.SeriesDate[6:8]+'.'+ref_PT.SeriesDate[4:6]+'.'+ref_PT.SeriesDate[0:4], \
                     'series_time': ref_PT.SeriesTime[0:2] + ref_PT.SeriesTime[2:4] + ref_PT.SeriesTime[4:6], \
                     'radionuclide': ref_PT[0x0054,0x0016][0][0x0054, 0x0300][0][0x0008,0x0104].value[1::], \
                     'halflife': str(ref_PT[0x0054,0x0016][0][0x0018, 0x1075].value), \
                     'injected_activity': str(ref_PT[0x0054,0x0016][0][0x0018, 0x1074].value/1000000), \
                     'injection_date': str(ref_PT[0x0054,0x0016][0][0x0018, 0x1078].value[6:8])+ \
                                           str(ref_PT[0x0054,0x0016][0][0x0018, 0x1078].value[4:6])+ \
                                               str(ref_PT[0x0054,0x0016][0][0x0018, 0x1078].value[0:4]), \
                     'injection_time': str(ref_PT[0x0054,0x0016][0][0x0018, 0x1072].value[0:6]), \
                     'device': ref_PT[0x0008,0x1090].value}

        self.data_size['text'] = 'rows: ' + str(data_info['rows']) + '\n' \
                            'columns: ' + str(data_info['cols']) + '\n' \
                            'slices: ' + str(data_info['slices']) + '\n' \
                            'slices thickness: ' + data_info['slice_thickness'] + ' mm' +'\n' \
                            'pixel size: ' + data_info['pixel_size'] + ' mm' +'\n' \
                            'study date: ' + data_info['study_date'] + '\n' \
                            'series date: ' + data_info['series_date'] + '\n' \
                            'series time: ' + data_info['series_time'] + '\n' \
                            'radionuclide: ' + data_info['radionuclide'] + '\n' \
                            'halflife: ' + data_info['halflife'] + ' seconds' + '\n' \
                            'device: ' + data_info['device']

        # sort the dictionary of slices according to the dictionary key
        temp_items = temp_data.items()
        temp_data = sorted(temp_items)

        # create 3D matrix - dictionary cannot be sliced in different planes
        PT_temp = np.zeros((data_info['rows'], data_info['cols'], data_info['slices']), dtype=int)
        i = 0
        for x in temp_data:
            PT_temp[:,:,i] = x[1]

            i += 1
# =============================================================================
# # add distorted pixels + smoothing
# # =============================================================================
        # # add distorted pixels
        # from skimage.draw import disk
        # mask = np.zeros((256, 256), dtype=np.uint8)
        # row = 120
        # col = 110
        # radius = 8
        # rr, cc = disk((row, col), radius)
        # mask[rr, cc] = 1
        # PT_temp[:,:,86][mask == 1] = 1.05*PT_temp[:,:,86][mask ==1]
        
        # # add more filtration
        # from math import pi, sqrt, exp

        # def gauss(n,sigma):
        #     r = range(-int(n/2),int(n/2)+1)
        #     return [1 / (sigma * sqrt(2*pi)) * exp(-float(x)**2/(2*sigma**2)) for x in r]

        # # FWHM filter
        # mm = 10
        # fwhm = mm/ref_PT.PixelSpacing[0]
        # # filter lenght
        # w = np.ceil(3*fwhm)
        # if w < 3:
        #     w = 3
                
        # print(fwhm, w)
        # g=np.array(gauss(w,fwhm)) # convert list to np array
        # print(g)
        # print(sum(g))
        # g = g/sum(g) # proper weighting
        # print(g, sum(g))

        # for r in range(0,data_info['rows']):
        #     PT_temp[r,:,86] = np.convolve(PT_temp[r,:,86], g, 'same')
        # for c in range(0, data_info['cols']):
        #     PT_temp[:,c,86] = np.convolve(PT_temp[:,c,86], g, 'same')
# # =============================================================================
# # normal function continues
# =============================================================================


        # check - additional filtration DICOM tag not present with Philips
        if ref_PT.get([0x0018,0x1210]) == None:
            ...
        else:
            data_info['filtration'] = ref_PT[0x0018,0x1210].value
            self.data_size['text'] = self.data_size['text'] + '\n' + \
                'Post-recon filter FWHM: ' + data_info['filtration'] + ' mm'

        # record data into phantom data - phantom.data (working), phantom.data_original (original volume activity)
        self.phantom = phantomData(dirname, PT_temp, ref_PT, [data_info['rows'], data_info['cols'], data_info['slices']], data_info)

        # primary setup of all variables
        self.primary_setup()

        # recalculate the values to SUV
        self.phantom.data = self.phantom.slice_in_SUV(self.phantom.data_original, self.data_volume.get().replace(',','.'), self.data_activity.get().replace(',','.'), self.data_time.get(), self.data_date.get())

        # find phantom's endpoint slices
        self.phantom.ending_position = self.phantom.find_me_phantom(self.phantom.uni_mask(self.phantom, float(self.phantom.uni_end)/10), self.ROI_extent_shrink.get())
        self.info_length = str((self.phantom.ending_position[1]-self.phantom.ending_position[0])*float(self.phantom.all_setup['slice_thickness']))

        # intial helper mask creation + thresholding - == 1 for the whole volume, threshold 0.1*max value
        self.phantom.mask = self.phantom.uni_mask(self.phantom, float(self.uni_center_lbl['text'])/100)

        # display the phantom -  self.phantom.show_me_slices((self, of_what, which, where, position, zoom))
        self.phantom.show_me_slices(self.phantom.data, self.phantom.slice, self.phantom_slice_view, self.phantom.position, \
                                    self.phantom.zoom, self.phantom.info[0], self.phantom.uni_thresh/100, self.ROI_results_show.get())

        #create the ROI for the evaluation - ROI parameters come from the central slice only
        self.uni_roi(self.phantom, int(0.5*(self.phantom.ending_position[1]+self.phantom.ending_position[0])), self.phantom_slice_view, self.phantom.uni_shrink)

        # setup the displayed slice label
        self.slice_info_updater()

        # =============================================================================
        # turn the frame_d on
        # =============================================================================
        frame_switch(self.frame_d, 1)
        # =============================================================================
        # bind to mouse events
        # =============================================================================
        self.phantom_slice_view.bind("<Shift-MouseWheel>", self.mouse_zoom)
        self.phantom_slice_view.bind("<MouseWheel>", self.mouse_list)
        self.phantom_slice_view.bind("<ButtonPress-2>", self.mouse_click_reset)

    # =============================================================================
    # mainWindow functions section
    # =============================================================================
    def slice_info_updater(self):
        slice_nr = self.phantom.slice
        textLU = 'slice: ' + str(slice_nr) + '\n' + \
                'mean SUV: ???' + '\n' + \
                'diff SUV (%): ???' + '\n' + \
                'COV (%): ???'
        textLB = 'SUV max: ???' + '\n' + \
                'SUV min: ???'
        textRB = 'ROI diameter: 000.0' + ' mm' +  '\n' + \
                'VOI length: 000.0' + ' mm'
        textRB = 'ROI diameter: ' + self.info_diameter + ' mm' +  '\n' + \
                'VOI length: ' + self.info_length + ' mm'

        if (self.phantom.slice_SUVmean != []) and \
            (slice_nr >= self.phantom.ending_position[0]) and (slice_nr <= self.phantom.ending_position[1]):
            textLU = 'slice: ' + str(slice_nr) + '\n' + \
                'mean SUV: ' + str(np.round(self.phantom.slice_SUVmean[slice_nr-self.phantom.ending_position[0]],3)) + '\n' + \
                'diff SUV (%): ' + str(np.round(self.phantom.slice_diff_from_1[slice_nr-self.phantom.ending_position[0]],3)) + '\n' + \
                'COV (%): ' + str(np.round(self.phantom.slice_COV[slice_nr-self.phantom.ending_position[0]],3))
            textLB = 'SUV max: ' + str(np.round(self.phantom.slice_SUVmax[slice_nr-self.phantom.ending_position[0]],3)) + '\n' + \
                'SUV min: ' + str(np.round(self.phantom.slice_SUVmin[slice_nr-self.phantom.ending_position[0]],3))
            textRB = 'ROI diameter: ' + self.info_diameter + ' mm' +  '\n' + \
                'VOI length: ' + self.info_length + ' mm'
        else:
            ...
        self.slice_infoLU.set(textLU)
        self.slice_infoLB.set(textLB)
        self.slice_infoRB.set(textRB)


    def primary_setup(self):
        self.uni_shrink_lbl['text'] = '0.0 / 0'
        self.uni_center_lbl['text'] = '50'
        self.uni_center_lbl['text'] = '50'
        self.uni_end_lbl['text'] = self.phantom.uni_end*10
        self.uni_thresh_lbl['text'] = self.phantom.uni_thresh
        self.uni_graph = 0
        self.uni_graph_crop_lbl['text'] = str(self.uni_graph)
        self.data_activity.delete(0, 'end')
        self.data_activity.insert(0, self.phantom.all_setup['injected_activity'])
        self.data_time.delete(0, 'end')
        self.data_time.insert(0, self.phantom.all_setup['injection_time'])
        self.data_date.delete(0, 'end')
        self.data_date.insert(0, self.phantom.all_setup['injection_date'])


    def mouse_zoom(self, event, *kwargs):
        zoom = [event.delta, event.x, event.y]
        self.phantom.slice_zoom(self, self.phantom.data, zoom, self.phantom_slice_view, self.ROI_results_show.get())
        if (self.phantom.slice >= self.phantom.ending_position[0]) and (self.phantom.slice <= self.phantom.ending_position[1]):
            self.uni_roi(self.phantom, int(0.5*(self.phantom.ending_position[1]+self.phantom.ending_position[0])), self.phantom_slice_view, self.phantom.uni_shrink)
        else:
            ...

    def mouse_list(self, event, *kwargs):
        index = event.delta
        self.phantom.slice_list(self, self.phantom.data, index, self.phantom_slice_view, self.ROI_results_show.get())
        self.slice_info_updater()
        if (self.phantom.slice >= self.phantom.ending_position[0]) and (self.phantom.slice <= self.phantom.ending_position[1]):
            self.uni_roi(self.phantom, int(0.5*(self.phantom.ending_position[1]+self.phantom.ending_position[0])), self.phantom_slice_view, self.phantom.uni_shrink)
        else:
            ...

    # reset button
    def mouse_click_reset(self, event):
        self.phantom.show_me_slices(self.phantom.data, self.phantom.slice, self.phantom_slice_view, [0,0], 1, self.phantom.info[0], self.phantom.uni_thresh/100, self.ROI_results_show.get())
        self.phantom.position=[0,0]
        self.phantom.zoom = 1
        if (self.phantom.slice >= self.phantom.ending_position[0]) and (self.phantom.slice <= self.phantom.ending_position[1]):
            self.uni_roi(self.phantom, int(0.5*(self.phantom.ending_position[1]+self.phantom.ending_position[0])), self.phantom_slice_view, self.phantom.uni_shrink)
        else:
            ...

    # changing the threshold for outliers
    def uni_thresh(self, i):
        self.phantom.uni_thresh += i

        if self.phantom.uni_thresh <= 1: self.phantom.uni_thresh = 1
        elif self.phantom.uni_thresh >= 100: self.phantom.uni_thresh = 100

        self.uni_thresh_lbl['text'] = str(self.phantom.uni_thresh)

    def uni_roi(self, what, which, where, diff):

        # also changing the length of the phantom for SUV / cut endpoint values
        self.phantom.ending_position = self.phantom.find_me_phantom(self.phantom.uni_mask(self.phantom, float(self.phantom.uni_end)/10), self.ROI_extent_shrink.get())

        where.delete('roi')
        # find the mask diameter
        # diff = shrink the mask by a pixel
        r_diam = float(np.count_nonzero(np.sum(what.mask[:,:,which], 0))+2*diff)
        c_diam = float(np.count_nonzero(np.sum(what.mask[:,:,which], 1))+2*diff)

        diameter = np.round(0.5*(r_diam*float(what.all_setup['pixel_size'])+c_diam*float(what.all_setup['pixel_size'])),1)

        self.info_diameter = str(diameter)

        center_row, center_col = what.give_me_center(what.data, which, what.size)
        x0 = center_col+c_diam/2
        y0 = center_row+r_diam/2
        x1 = center_col-c_diam/2
        y1 = center_row-r_diam/2

        x0, y0, s0 = what.slice_click_reversed(self, (x0, y0, which))

        x1, y1, s1 = what.slice_click_reversed(self, (x1, y1, which))

        where.create_oval(y0, x0, y1, x1, outline='yellow', width=2, tags='roi')

        self.slice_info_updater()

    def uni_shrink(self, i):
        self.phantom.uni_shrink += i

        if self.phantom.uni_shrink <= -20: self.phantom.uni_shrink = -20
        elif self.phantom.uni_shrink >= 0: self.phantom.uni_shrink = 0

        diff = np.round(self.phantom.uni_shrink*float(self.phantom.all_setup['pixel_size']),2)

        self.uni_shrink_lbl['text'] = str(diff) + ' / ' + str(self.phantom.uni_shrink)

        self.uni_roi(self.phantom, int(0.5*(self.phantom.ending_position[1]+self.phantom.ending_position[0])), self.phantom_slice_view, self.phantom.uni_shrink)

        self.slice_info_updater()

    def uni_center(self, i):
        self.phantom.uni_center += 0.5*i

        if self.phantom.uni_center <= 0: self.phantom.uni_center = 0
        elif self.phantom.uni_center >= 9.5: self.phantom.uni_center = 9.5

        self.uni_center_lbl['text'] = str(int(self.phantom.uni_center*10))

        # create a helper mask + thresholding the mask - == 1 for the whole phantom volume # threshold 0.1 multiple of the max
        self.phantom.mask = self.phantom.uni_mask(self.phantom, self.phantom.uni_center/10)

        self.uni_roi(self.phantom, int(0.5*(self.phantom.ending_position[1]+self.phantom.ending_position[0])), self.phantom_slice_view, self.phantom.uni_shrink)

        self.slice_info_updater()

    def uni_end(self, i):
        self.phantom.uni_end += 0.5*i

        if self.phantom.uni_end <= 0: self.phantom.uni_end = 0
        elif self.phantom.uni_end >= 9.5: self.phantom.uni_end = 9.5

        self.uni_end_lbl['text'] = str(int(self.phantom.uni_end*10))

        # locate endpoint slices of the phantom
        self.phantom.ending_position = self.phantom.find_me_phantom(self.phantom.uni_mask(self.phantom, self.phantom.uni_end/10), self.ROI_extent_shrink.get())

        self.info_length = str((self.phantom.ending_position[1]-self.phantom.ending_position[0])*float(self.phantom.all_setup['slice_thickness']))

        self.uni_roi(self.phantom, int(0.5*(self.phantom.ending_position[1]+self.phantom.ending_position[0])), self.phantom_slice_view, self.phantom.uni_shrink)

        self.slice_info_updater()

    # change the extent of the endpoint slices to be cut out for the graph display
    def uni_graph_crop(self, i):
        self.uni_graph += i

        if self.uni_graph <= 0: self.uni_graph = 0
        elif self.uni_graph >= 10: self.uni_graph = 10

        self.uni_graph_crop_lbl['text'] = str(self.uni_graph)

        # x axis preparation step
        if self.phantom.slice_diff_from_1 != []:
            slice_nr = list(range(self.phantom.ending_position[0]+self.uni_graph, self.phantom.ending_position[1]+1-self.uni_graph))
            if self.uni_graph !=0:
                slice_diff_from_1 = self.phantom.slice_diff_from_1[self.uni_graph:-self.uni_graph]
                slice_COV = self.phantom.slice_COV[self.uni_graph:-self.uni_graph]
            else:
                slice_diff_from_1 = self.phantom.slice_diff_from_1
                slice_COV = self.phantom.slice_COV
            GraphDiffCOV(self.frame_b, (self.Xsize, self.Ysize), [0,0], [slice_nr,slice_diff_from_1], [slice_nr,slice_COV])
        else:
            ...

    def calculate_SUV(self):
        # locate endpoint slices of the phantom
        self.phantom.ending_position = self.phantom.find_me_phantom(self.phantom.uni_mask(self.phantom, self.phantom.uni_end/10), self.ROI_extent_shrink.get())
        self.uni_graph = 0 # set graph cut to the initial value - no cut, full extent
        self.uni_graph_crop_lbl['text'] = str(self.uni_graph)

        self.phantom.data = self.phantom.slice_in_SUV(self.phantom.data_original, self.data_volume.get().replace(',','.'), self.data_activity.get().replace(',','.'), self.data_time.get(), self.data_date.get())
        self.phantom.info[0], self.phantom.info[1], self.phantom.SUV_mask = self.phantom.give_me_SUV(self.phantom, int(self.uni_center_lbl['text'])/100, self.phantom.uni_shrink)
        self.info_SUV.set(self.phantom.info[0])
        self.info_SUV_deviation.set(self.phantom.info[1])
        self.phantom.show_me_slices(self.phantom.data, self.phantom.slice, self.phantom_slice_view, self.phantom.position, \
                            self.phantom.zoom, self.phantom.info[0], self.phantom.uni_thresh/100, self.ROI_results_show.get())
        self.uni_roi(self.phantom, self.phantom.slice, self.phantom_slice_view, self.phantom.uni_shrink)
        # display the graph
        # x axis preparation step
        slice_nr = list(range(self.phantom.ending_position[0], self.phantom.ending_position[1]+1))
        GraphDiffCOV(self.frame_b, (self.Xsize, self.Ysize), [0,0], [slice_nr,self.phantom.slice_diff_from_1], [slice_nr,self.phantom.slice_COV])

        self.slice_info_updater()

    def ROI_results_hide(self):
        self.phantom.show_me_slices(self.phantom.data, self.phantom.slice, self.phantom_slice_view, self.phantom.position, \
                    self.phantom.zoom, self.phantom.info[0], self.phantom.uni_thresh/100, self.ROI_results_show.get())
        self.uni_roi(self.phantom, int(0.5*(self.phantom.ending_position[1]+self.phantom.ending_position[0])), self.phantom_slice_view, self.phantom.uni_shrink)

    def get_out_results(self):
        self.phantom.get_out_results()

#%% global functions
# function for switching frames on and off
def frame_switch(frame_name, on_off):
    if on_off == 0:
        for child in frame_name.winfo_children():
            child.configure(state='disabled')
    else:
        for child in frame_name.winfo_children():
            child.configure(state='normal')

root = tk.Tk()
uniformity_analyzer = mainWindow(root)
root.mainloop()